package debixApiSampleCode;

import java.security.KeyPair;
import java.security.NoSuchAlgorithmException;
import java.security.cert.X509Certificate;
import java.security.interfaces.ECPublicKey;
import java.security.spec.InvalidKeySpecException;
import java.util.Base64;
import java.util.Date;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.nimbusds.jose.JOSEException;
import com.nimbusds.jose.JWSHeader;
import com.nimbusds.jose.Payload;
import com.nimbusds.jose.crypto.RSASSASigner;
import com.nimbusds.jwt.SignedJWT;
import ch.six.cards.library.debix.bankapi.bank_to_six.v2.dto.CardCredentialsRequestJwsPayload;
import ch.six.cards.library.debix.bankapi.bank_to_six.v2.dto.CardCredentialsResponseJwsPayload;
import ch.six.cards.library.debix.bankapi.bank_to_six.v2.dto.CardCredentialsResponseUnencryptedData;
import static com.nimbusds.jose.JWSAlgorithm.RS256;
import static com.nimbusds.jose.util.X509CertUtils.computeSHA256Thumbprint;
import static debixApiSampleCode.JwsEncryptionUtil.ELLIPTIC_CURVE_ALGORITHM;
import static debixApiSampleCode.JwsEncryptionUtil.PRIME_256_V1_PARAMETER_SPEC;
import static debixApiSampleCode.JwsEncryptionUtil.createJwsHeaderWith;
import static debixApiSampleCode.JwsEncryptionUtil.encrypt;
import static debixApiSampleCode.JwsEncryptionUtil.extractAndValidateJwsRequest;
import static debixApiSampleCode.JwsEncryptionUtil.generateKeyPair;
import static debixApiSampleCode.JwsEncryptionUtil.generateSelfSignedX509Certificate;
import static debixApiSampleCode.JwsEncryptionUtil.generateSharedSecret;
import static debixApiSampleCode.JwsEncryptionUtil.jwtClaimsSetFrom;
import static debixApiSampleCode.JwsEncryptionUtil.toEcPublicKey;
import static debixApiSampleCode.JwsEncryptionUtil.toX509EncodedEcPublicKey;

class DebixApiGetCardCredentialsSampleCodeForDebix {

  private static final String RSA_SIGNING_ALGORITHM = "RSA";
  private static final String DEBIX_CERTIFICATE_SIGNING_ALGORITHM = "SHA256withRSA";
  private static final ObjectMapper OBJECT_MAPPER = new ObjectMapper();

  private final KeyPair debixEphemeralKeyPair;
  private final X509Certificate debixSigningCertificate;
  private final KeyPair debixSigningKeyPair;

  public DebixApiGetCardCredentialsSampleCodeForDebix() {
    this.debixSigningKeyPair = generateKeyPair(RSA_SIGNING_ALGORITHM, 2048);
    // this certificate needs to be given to the issuer in advance, so the issuer can validate the signature
    this.debixSigningCertificate = generateSelfSignedX509Certificate(
        this.debixSigningKeyPair,
        "cn=debix.test",
        DEBIX_CERTIFICATE_SIGNING_ALGORITHM,
        new Date(),
        730
    );

    // generated new, for each card credential call
    this.debixEphemeralKeyPair = generateKeyPair(ELLIPTIC_CURVE_ALGORITHM, PRIME_256_V1_PARAMETER_SPEC);
  }

  public String getCardCredentials(String cardCredentialsClientJwsRequest, X509Certificate clientSigningCertificate) {
    CardCredentialsRequestJwsPayload cardCredentialsRequestJwsPayload = extractAndValidateJwsRequest(
        clientSigningCertificate,
        cardCredentialsClientJwsRequest,
        CardCredentialsRequestJwsPayload.class
    );
    CardCredentialsResponseJwsPayload cardCredentialsResponseJwsPayload = generateCardCredentialsResponse(cardCredentialsRequestJwsPayload);

    return encodeAndSignCardCredentialsResponseJwt(cardCredentialsResponseJwsPayload);
  }

  public X509Certificate getSigningCertificate() {
    return this.debixSigningCertificate;
  }

  private CardCredentialsResponseJwsPayload generateCardCredentialsResponse(CardCredentialsRequestJwsPayload cardCredentialsRequestJwsPayload) {
    byte[] x509EncodedEphemeralDebixPublicKey;
    try {
      x509EncodedEphemeralDebixPublicKey = toX509EncodedEcPublicKey(this.debixEphemeralKeyPair.getPublic());
    } catch (NoSuchAlgorithmException | InvalidKeySpecException e) {
      throw new RuntimeException("failed to X509-encode an EC public key", e);
    }
    try {
      ECPublicKey ephemeralClientPublicKey = toEcPublicKey(Base64.getDecoder().decode(cardCredentialsRequestJwsPayload.getClientEphPubKey()));
      byte[] sharedSecret = generateSharedSecret(this.debixEphemeralKeyPair.getPrivate(), ephemeralClientPublicKey);

      CardCredentialsResponseUnencryptedData cardCredentialsResponseUnencryptedData = new CardCredentialsResponseUnencryptedData()
          .fpan("1234567890123456")
          .cvv("111");
      byte[] unencryptedData = OBJECT_MAPPER.writeValueAsBytes(cardCredentialsResponseUnencryptedData);
      String encryptedData = Base64.getEncoder().encodeToString(encrypt(sharedSecret, unencryptedData));

      return new CardCredentialsResponseJwsPayload()
          .serverEphPubKey(Base64.getEncoder().encodeToString(x509EncodedEphemeralDebixPublicKey))
          .encryptedData(encryptedData);
    } catch (Exception e) {
      throw new RuntimeException(e);
    }
  }

  private String encodeAndSignCardCredentialsResponseJwt(CardCredentialsResponseJwsPayload cardCredentialsResponseJwsPayload) {
    Payload payload;
    try {
      payload = new Payload(OBJECT_MAPPER.writeValueAsBytes(cardCredentialsResponseJwsPayload));
    } catch (JsonProcessingException e) {
      throw new RuntimeException("Failed to convert response JWS payload: " + e.getMessage());
    }
    JWSHeader jwsHeader = createJwsHeaderWith(RS256, computeSHA256Thumbprint(this.debixSigningCertificate));
    SignedJWT signedJwt = new SignedJWT(jwsHeader, jwtClaimsSetFrom(payload.toJSONObject()));
    try {
      RSASSASigner jwsSigner = new RSASSASigner(this.debixSigningKeyPair.getPrivate());
      signedJwt.sign(jwsSigner);
    } catch (JOSEException e) {
      throw new RuntimeException("Failed to sign the JWT. -> " + e.getMessage(), e);
    }

    return signedJwt.serialize();
  }

}

