package debixApiSampleCode;

import java.security.KeyPair;
import java.security.NoSuchAlgorithmException;
import java.security.cert.X509Certificate;
import java.security.interfaces.ECPrivateKey;
import java.security.interfaces.ECPublicKey;
import java.security.spec.InvalidKeySpecException;
import java.util.Base64;
import java.util.Date;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.nimbusds.jose.JOSEException;
import com.nimbusds.jose.JWSHeader;
import com.nimbusds.jose.Payload;
import com.nimbusds.jose.crypto.ECDSASigner;
import com.nimbusds.jwt.SignedJWT;
import ch.six.cards.library.debix.bankapi.bank_to_six.v2.dto.CardId;
import ch.six.cards.library.debix.bankapi.bank_to_six.v2.dto.GetPinRequestJwsPayload;
import ch.six.cards.library.debix.bankapi.bank_to_six.v2.dto.GetPinResponseJwsPayload;
import ch.six.cards.library.debix.bankapi.bank_to_six.v2.dto.GetPinResponseUnencryptedData;
import ch.six.cards.library.debix.bankapi.bank_to_six.v2.dto.YearMonth;
import static com.nimbusds.jose.JWSAlgorithm.ES256;
import static com.nimbusds.jose.util.X509CertUtils.computeSHA256Thumbprint;
import static debixApiSampleCode.JwsEncryptionUtil.ELLIPTIC_CURVE_ALGORITHM;
import static debixApiSampleCode.JwsEncryptionUtil.PRIME_256_V1_PARAMETER_SPEC;
import static debixApiSampleCode.JwsEncryptionUtil.SHA256_WITH_ECDSA_SIGNING_ALGORITHM;
import static debixApiSampleCode.JwsEncryptionUtil.createJwsHeaderWith;
import static debixApiSampleCode.JwsEncryptionUtil.decrypt;
import static debixApiSampleCode.JwsEncryptionUtil.extractAndValidateJwsRequest;
import static debixApiSampleCode.JwsEncryptionUtil.generateKeyPair;
import static debixApiSampleCode.JwsEncryptionUtil.generateSelfSignedX509Certificate;
import static debixApiSampleCode.JwsEncryptionUtil.generateSharedSecret;
import static debixApiSampleCode.JwsEncryptionUtil.jwtClaimsSetFrom;
import static debixApiSampleCode.JwsEncryptionUtil.toEcPublicKey;
import static debixApiSampleCode.JwsEncryptionUtil.toX509EncodedEcPublicKey;

class DebixApiGetPinSampleCodeForIssuer {

  private static final Logger LOGGER = LoggerFactory.getLogger(DebixApiGetPinSampleCodeForIssuer.class);
  private static final ObjectMapper OBJECT_MAPPER = new ObjectMapper();

  private DebixApiGetPinSampleCodeForIssuer() {
  }

  public static void main(String[] args) {
    KeyPair issuerSigningKeyPair = generateKeyPair(ELLIPTIC_CURVE_ALGORITHM, PRIME_256_V1_PARAMETER_SPEC);
    // NOTE: this certificate needs to be given to debiX in advance, so SIX can validate the signature
    X509Certificate issuerCertificate = generateSelfSignedX509Certificate(
        issuerSigningKeyPair,
        "cn=issuer.test",
        SHA256_WITH_ECDSA_SIGNING_ALGORITHM,
        new Date(),
        730
    );

    CardId cardId = new CardId()
        .bankClearingNumber(4711)
        .cardNumber(12345)
        .cardType(1)
        .cardExpiry(new YearMonth().month("10").year("25"))
        .cardSeqNumber(1);

    // Issuer get pin request
    KeyPair issuerEphemeralKeyPair = generateKeyPair(ELLIPTIC_CURVE_ALGORITHM, PRIME_256_V1_PARAMETER_SPEC);
    String getPinClientJwsRequest = generateGetPinClientJwsRequest(issuerSigningKeyPair, issuerCertificate, issuerEphemeralKeyPair, cardId);

    // debiX start get pin
    DebixApiGetPinSampleCodeForDebix debix = new DebixApiGetPinSampleCodeForDebix();
    String getPinResponseFromDebix = debix.getPin(getPinClientJwsRequest, issuerCertificate);

    // NOTE: this certificate must already be present in the issuer system, so the issuer can verify the debiX response
    X509Certificate debixSigningCertificate = debix.getSigningCertificate();
    GetPinResponseJwsPayload getPinResponseJwsPayload = extractAndValidateJwsRequest(
        debixSigningCertificate,
        getPinResponseFromDebix,
        GetPinResponseJwsPayload.class
    );
    decryptGetPin(issuerEphemeralKeyPair, getPinResponseJwsPayload);
  }

  public static String generateGetPinClientJwsRequest(
      KeyPair clientSigningKeyPair,
      X509Certificate clientSigningCertificate,
      KeyPair issuerEphemeralKeyPair,
      CardId cardId
  ) {
    JWSHeader requestJwsHeader = createJwsHeaderWith(ES256, computeSHA256Thumbprint(clientSigningCertificate), "keyId");

    SignedJWT signedRequestJwt;
    try {
      String clientEphemeralPublicKeyPemBase64 = Base64.getEncoder().encodeToString(toX509EncodedEcPublicKey(issuerEphemeralKeyPair.getPublic()));
      GetPinRequestJwsPayload getPinRequestJwsPayload = new GetPinRequestJwsPayload()
          .cardId(cardId)
          .clientEphPubKey(clientEphemeralPublicKeyPemBase64);
      Payload payload = new Payload(OBJECT_MAPPER.writeValueAsBytes(getPinRequestJwsPayload));
      signedRequestJwt = new SignedJWT(requestJwsHeader, jwtClaimsSetFrom(payload.toJSONObject()));
      signedRequestJwt.sign(new ECDSASigner((ECPrivateKey) clientSigningKeyPair.getPrivate()));
    } catch (JsonProcessingException | JOSEException | NoSuchAlgorithmException | InvalidKeySpecException e) {
      throw new RuntimeException("Failed to generate JWS.", e);
    }

    return signedRequestJwt.serialize();
  }

  private static void decryptGetPin(KeyPair issuerEphemeralKeyPair, GetPinResponseJwsPayload getPinResponseJwsPayload) {
    try {
      ECPublicKey serverEphemeralPublicKey = toEcPublicKey(Base64.getDecoder().decode(getPinResponseJwsPayload.getServerEphPubKey()));
      byte[] sharedSecret = generateSharedSecret(serverEphemeralPublicKey, issuerEphemeralKeyPair.getPrivate(), serverEphemeralPublicKey);
      byte[] decryptedData = decrypt(sharedSecret, Base64.getDecoder().decode(getPinResponseJwsPayload.getEncryptedData()));

      GetPinResponseUnencryptedData getPinResponseUnencryptedData = OBJECT_MAPPER.readValue(decryptedData, GetPinResponseUnencryptedData.class);
      LOGGER.info("Issuer:");
      LOGGER.info("Got pin: \"{}\"", getPinResponseUnencryptedData.getPin());
    } catch (Exception e) {
      throw new RuntimeException(e);
    }
  }

}
