package debixApiSampleCode;

import java.security.KeyPair;
import java.security.cert.X509Certificate;
import java.security.interfaces.ECPrivateKey;
import java.security.interfaces.ECPublicKey;
import java.util.Base64;
import java.util.Date;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.nimbusds.jose.JOSEException;
import com.nimbusds.jose.JWSHeader;
import com.nimbusds.jose.Payload;
import com.nimbusds.jose.crypto.ECDSASigner;
import com.nimbusds.jwt.SignedJWT;
import ch.six.cards.library.debix.bankapi.bank_to_six.v2.dto.CardId;
import ch.six.cards.library.debix.bankapi.bank_to_six.v2.dto.SetPinRequestJwsPayload;
import ch.six.cards.library.debix.bankapi.bank_to_six.v2.dto.SetPinRequestUnencryptedData;
import ch.six.cards.library.debix.bankapi.bank_to_six.v2.dto.StartSetPinRequestJwsPayload;
import ch.six.cards.library.debix.bankapi.bank_to_six.v2.dto.StartSetPinResponseJwsPayload;
import ch.six.cards.library.debix.bankapi.bank_to_six.v2.dto.YearMonth;
import static com.nimbusds.jose.JWSAlgorithm.ES256;
import static com.nimbusds.jose.util.X509CertUtils.computeSHA256Thumbprint;
import static debixApiSampleCode.JwsEncryptionUtil.ELLIPTIC_CURVE_ALGORITHM;
import static debixApiSampleCode.JwsEncryptionUtil.PRIME_256_V1_PARAMETER_SPEC;
import static debixApiSampleCode.JwsEncryptionUtil.createJwsHeaderWith;
import static debixApiSampleCode.JwsEncryptionUtil.encrypt;
import static debixApiSampleCode.JwsEncryptionUtil.extractAndValidateJwsRequest;
import static debixApiSampleCode.JwsEncryptionUtil.generateKeyPair;
import static debixApiSampleCode.JwsEncryptionUtil.generateSelfSignedX509Certificate;
import static debixApiSampleCode.JwsEncryptionUtil.generateSharedSecret;
import static debixApiSampleCode.JwsEncryptionUtil.jwtClaimsSetFrom;
import static debixApiSampleCode.JwsEncryptionUtil.toEcPublicKey;
import static debixApiSampleCode.JwsEncryptionUtil.toX509EncodedEcPublicKey;

class DebixApiSetPinSampleCodeForIssuer {

  private static final ObjectMapper OBJECT_MAPPER = new ObjectMapper();

  private DebixApiSetPinSampleCodeForIssuer() {
  }

  public static void main(String[] args) {
    KeyPair issuerSigningKeyPair = generateKeyPair(ELLIPTIC_CURVE_ALGORITHM, PRIME_256_V1_PARAMETER_SPEC);
    // NOTE: this certificate needs to be given to debiX in advance, so SIX can validate the signature
    X509Certificate issuerCertificate = generateSelfSignedX509Certificate(
        issuerSigningKeyPair,
        "cn=issuer.test",
        JwsEncryptionUtil.SHA256_WITH_ECDSA_SIGNING_ALGORITHM,
        new Date(),
        730
    );

    CardId cardId = new CardId()
        .bankClearingNumber(4711)
        .cardNumber(12345)
        .cardType(1)
        .cardExpiry(new YearMonth().month("10").year("25"))
        .cardSeqNumber(1);

    // Issuer start set pin request
    String startSetPinClientJwsRequest = generateStartSetPinClientJwsRequest(issuerSigningKeyPair, issuerCertificate, cardId);

    // debiX start set pin
    DebixApiSetPinSampleCodeForDebix debix = new DebixApiSetPinSampleCodeForDebix();
    String startSetPinResponseFromDebix = debix.startSetPin(startSetPinClientJwsRequest, issuerCertificate);

    // NOTE: this certificate must already be present in the issuer system, so the issuer can verify the debiX response
    X509Certificate debixSigningCertificate = debix.getSigningCertificate();
    StartSetPinResponseJwsPayload startSetPinResponseJwsPayload = extractAndValidateJwsRequest(
        debixSigningCertificate,
        startSetPinResponseFromDebix,
        StartSetPinResponseJwsPayload.class
    );

    // issuer set pin request
    KeyPair issuerEphemeralKeyPair = generateKeyPair(ELLIPTIC_CURVE_ALGORITHM, PRIME_256_V1_PARAMETER_SPEC);
    String issuerSetPinJwsRequest = generateSetPinClientJws(cardId, issuerSigningKeyPair, issuerCertificate, issuerEphemeralKeyPair, startSetPinResponseJwsPayload);

    // debiX set pin
    debix.setPin(issuerSetPinJwsRequest, issuerCertificate);
  }

  public static String generateStartSetPinClientJwsRequest(KeyPair clientSigningKeyPair, X509Certificate clientSigningCertificate, CardId cardId) {
    JWSHeader requestJwsHeader = createJwsHeaderWith(ES256, computeSHA256Thumbprint(clientSigningCertificate), "keyId");

    SignedJWT signedRequestJwt;
    try {
      StartSetPinRequestJwsPayload startSetPinRequestJwsPayload = new StartSetPinRequestJwsPayload()
          .cardId(cardId);
      Payload payload = new Payload(OBJECT_MAPPER.writeValueAsBytes(startSetPinRequestJwsPayload));
      signedRequestJwt = new SignedJWT(requestJwsHeader, jwtClaimsSetFrom(payload.toJSONObject()));
      signedRequestJwt.sign(new ECDSASigner((ECPrivateKey) clientSigningKeyPair.getPrivate()));
    } catch (JsonProcessingException | JOSEException e) {
      throw new RuntimeException("Failed to generate JWS.", e);
    }

    return signedRequestJwt.serialize();
  }

  public static String generateSetPinClientJws(
      CardId cardId,
      KeyPair issuerSigningKeyPair,
      X509Certificate issuerCertificate,
      KeyPair issuerEphemeralKeyPair,
      StartSetPinResponseJwsPayload startSetPinResponseJwsPayload
  ) {
    SetPinRequestJwsPayload setPinRequestJwsPayload;
    try {
      ECPublicKey serverPublicKey = toEcPublicKey(Base64.getDecoder().decode(startSetPinResponseJwsPayload.getServerEphPubKey()));
      byte[] sharedSecret = generateSharedSecret(serverPublicKey, issuerEphemeralKeyPair.getPrivate(), serverPublicKey);
      byte[] unencryptedData = OBJECT_MAPPER.writeValueAsBytes(new SetPinRequestUnencryptedData().pin("123456"));
      String encryptedData = Base64.getEncoder().encodeToString(encrypt(sharedSecret, unencryptedData));
      String clientEphemeralPublicKeyPemBase64 = Base64.getEncoder().encodeToString(toX509EncodedEcPublicKey(issuerEphemeralKeyPair.getPublic()));

      setPinRequestJwsPayload = new SetPinRequestJwsPayload()
          .cardId(cardId)
          .clientEphPubKey(clientEphemeralPublicKeyPemBase64)
          .encryptedData(encryptedData);
    } catch (Exception e) {
      throw new RuntimeException("Failed to parse and convert JWS.", e);
    }

    JWSHeader requestJwsHeader = createJwsHeaderWith(ES256, computeSHA256Thumbprint(issuerCertificate), "keyId");

    SignedJWT signedRequestJwt;
    try {
      Payload payload = new Payload(OBJECT_MAPPER.writeValueAsBytes(setPinRequestJwsPayload));
      signedRequestJwt = new SignedJWT(requestJwsHeader, jwtClaimsSetFrom(payload.toJSONObject()));
      signedRequestJwt.sign(new ECDSASigner((ECPrivateKey) issuerSigningKeyPair.getPrivate()));
    } catch (Exception e) {
      throw new RuntimeException("Failed to generate JWS.", e);
    }

    return signedRequestJwt.serialize();
  }

}
