package debixApiSampleCode;

import java.math.BigInteger;
import java.security.InvalidAlgorithmParameterException;
import java.security.InvalidKeyException;
import java.security.KeyFactory;
import java.security.KeyPair;
import java.security.KeyPairGenerator;
import java.security.NoSuchAlgorithmException;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.security.interfaces.ECPublicKey;
import java.security.interfaces.RSAPublicKey;
import java.security.spec.AlgorithmParameterSpec;
import java.security.spec.InvalidKeySpecException;
import java.security.spec.X509EncodedKeySpec;
import java.text.ParseException;
import java.time.ZonedDateTime;
import java.util.Date;
import java.util.Map;
import java.util.concurrent.TimeUnit;
import javax.crypto.BadPaddingException;
import javax.crypto.Cipher;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.KeyAgreement;
import javax.crypto.NoSuchPaddingException;
import javax.crypto.SecretKey;
import javax.crypto.spec.GCMParameterSpec;
import javax.crypto.spec.SecretKeySpec;
import org.bouncycastle.asn1.x500.X500Name;
import org.bouncycastle.asn1.x509.BasicConstraints;
import org.bouncycastle.asn1.x509.ExtendedKeyUsage;
import org.bouncycastle.asn1.x509.Extension;
import org.bouncycastle.asn1.x509.KeyPurposeId;
import org.bouncycastle.asn1.x509.KeyUsage;
import org.bouncycastle.asn1.x509.SubjectPublicKeyInfo;
import org.bouncycastle.cert.CertIOException;
import org.bouncycastle.cert.X509CertificateHolder;
import org.bouncycastle.cert.X509v3CertificateBuilder;
import org.bouncycastle.cert.jcajce.JcaX509CertificateConverter;
import org.bouncycastle.cert.jcajce.JcaX509ExtensionUtils;
import org.bouncycastle.jcajce.spec.UserKeyingMaterialSpec;
import org.bouncycastle.jce.provider.BouncyCastleProvider;
import org.bouncycastle.operator.ContentSigner;
import org.bouncycastle.operator.OperatorCreationException;
import org.bouncycastle.operator.jcajce.JcaContentSignerBuilder;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.nimbusds.jose.JOSEException;
import com.nimbusds.jose.JWSAlgorithm;
import com.nimbusds.jose.JWSHeader;
import com.nimbusds.jose.JWSVerifier;
import com.nimbusds.jose.crypto.ECDSAVerifier;
import com.nimbusds.jose.crypto.RSASSAVerifier;
import com.nimbusds.jose.util.Base64URL;
import com.nimbusds.jwt.JWTClaimsSet;
import com.nimbusds.jwt.SignedJWT;
import static com.nimbusds.jose.JOSEObjectType.JWT;
import static com.nimbusds.jose.util.X509CertUtils.computeSHA256Thumbprint;
import static java.time.ZoneId.systemDefault;
import static java.util.Arrays.copyOfRange;
import static javax.crypto.Cipher.DECRYPT_MODE;
import static javax.crypto.Cipher.ENCRYPT_MODE;
import static org.bouncycastle.jce.ECNamedCurveTable.getParameterSpec;

public final class JwsEncryptionUtil {

  static final String ELLIPTIC_CURVE_ALGORITHM = "EC";
  static final String SHA256_WITH_ECDSA_SIGNING_ALGORITHM = "SHA256withECDSA";
  static final String KEY_EXCHANGE_AND_DERIVATION_ALGORITHM = "ECCDHwithSHA256KDF";
  static final String PRIME_256_V1_PARAMETER_SPEC = "prime256v1";
  // Must only be initialized once, as it has a memory leak
  private static final BouncyCastleProvider BOUNCY_CASTLE_PROVIDER = new BouncyCastleProvider();
  private static final String AES_GCM_CIPHER_ALGORITHM = "AES/GCM/NoPadding";
  private static final String AES_KEY_ALGORITHM = "AES";
  private static final int AES_KEY_SHARED_SECRET_OFFSET = 0;
  private static final int AES_KEY_BYTE_LENGTH = 128 / 8;
  private static final int AES_IV_SHARED_SECRET_OFFSET = 128 / 8;
  private static final int AES_IV_BYTE_LENGTH = 128 / 8;
  private static final int AES_GCM_AUTHENTICATION_TAG_LENGTH = 128;
  private static final ObjectMapper OBJECT_MAPPER = new ObjectMapper();

  private JwsEncryptionUtil() {
  }

  static JWSHeader createJwsHeaderWith(JWSAlgorithm jwsAlgorithm, Base64URL sha256Thumbprint, String keyId) {
    return new JWSHeader.Builder(jwsAlgorithm)
        .type(JWT)
        .keyID(keyId)
        .x509CertSHA256Thumbprint(sha256Thumbprint)
        .build();
  }

  static ECPublicKey toEcPublicKey(byte[] x509Encoded) throws NoSuchAlgorithmException, InvalidKeySpecException {
    return (ECPublicKey) KeyFactory.getInstance(ELLIPTIC_CURVE_ALGORITHM)
        .generatePublic(new X509EncodedKeySpec(x509Encoded, ELLIPTIC_CURVE_ALGORITHM));
  }

  static byte[] toX509EncodedEcPublicKey(PublicKey publicKey) throws NoSuchAlgorithmException, InvalidKeySpecException {
    return KeyFactory.getInstance(publicKey.getAlgorithm())
        .getKeySpec(publicKey, X509EncodedKeySpec.class)
        .getEncoded();
  }

  static KeyPair generateKeyPair(String algorithm, String parameterSpec) {
    return generateKeyPair(algorithm, parameterSpec, -1);
  }

  static KeyPair generateKeyPair(String algorithm, int keySize) {
    return generateKeyPair(algorithm, null, keySize);
  }

  static <T> T extractAndValidateJwsRequest(X509Certificate signingCertificate, String rawRequestJws, Class<T> jwsPayloadClass) {
    SignedJWT parsedCardCredentialsRequestJws;
    try {
      parsedCardCredentialsRequestJws = SignedJWT.parse(rawRequestJws);
    } catch (ParseException e) {
      throw new RuntimeException("Failed to parse request JWS.", e);
    }

    validateRequestThumbprintAndSignature(parsedCardCredentialsRequestJws, signingCertificate);

    return OBJECT_MAPPER.convertValue(
        parsedCardCredentialsRequestJws.getPayload().toJSONObject(),
        jwsPayloadClass
    );
  }

  static byte[] generateSharedSecret(PrivateKey localPrivateKey, PublicKey remotePublicKey) throws NoSuchAlgorithmException, InvalidKeyException {
    KeyAgreement keyAgreement = KeyAgreement.getInstance(KEY_EXCHANGE_AND_DERIVATION_ALGORITHM, BOUNCY_CASTLE_PROVIDER);
    keyAgreement.init(localPrivateKey);
    keyAgreement.doPhase(remotePublicKey, true);

    return keyAgreement.generateSecret();
  }

  static byte[] generateSharedSecret(
      PublicKey localPublicKey,
      PrivateKey localPrivateKey,
      PublicKey remotePublicKey
  ) throws NoSuchAlgorithmException, InvalidKeyException, InvalidAlgorithmParameterException {
    KeyAgreement keyAgreement = KeyAgreement.getInstance(KEY_EXCHANGE_AND_DERIVATION_ALGORITHM, BOUNCY_CASTLE_PROVIDER);
    // used to prevent private key leakage see https://darthnull.org/secure-enclave-ecies/
    UserKeyingMaterialSpec userKeyingMaterialSpec = new UserKeyingMaterialSpec(SubjectPublicKeyInfo.getInstance(localPublicKey.getEncoded()).getPublicKeyData().getBytes());
    keyAgreement.init(localPrivateKey, userKeyingMaterialSpec);
    keyAgreement.doPhase(remotePublicKey, true);

    return keyAgreement.generateSecret();
  }

  static byte[] encrypt(
      byte[] sharedSecret,
      byte[] unencryptedData
  ) throws NoSuchPaddingException, NoSuchAlgorithmException, InvalidAlgorithmParameterException, InvalidKeyException, IllegalBlockSizeException, BadPaddingException {
    Cipher aesCipher = Cipher.getInstance(AES_GCM_CIPHER_ALGORITHM);
    SecretKey secretKey = new SecretKeySpec(extractAesKeyFromSharedSecret(sharedSecret), AES_KEY_ALGORITHM);
    GCMParameterSpec gcmParamSpec = new GCMParameterSpec(AES_GCM_AUTHENTICATION_TAG_LENGTH, extractAesIvFromSharedSecret(sharedSecret));
    aesCipher.init(ENCRYPT_MODE, secretKey, gcmParamSpec);

    return aesCipher.doFinal(unencryptedData);
  }

  static byte[] decrypt(
      byte[] sharedSecret,
      byte[] encryptedData
  ) throws NoSuchPaddingException, NoSuchAlgorithmException, InvalidAlgorithmParameterException, InvalidKeyException, IllegalBlockSizeException, BadPaddingException {
    Cipher aesCipher = Cipher.getInstance(AES_GCM_CIPHER_ALGORITHM);
    SecretKey secretKey = new SecretKeySpec(extractAesKeyFromSharedSecret(sharedSecret), AES_KEY_ALGORITHM);
    GCMParameterSpec gcmParamSpec = new GCMParameterSpec(AES_GCM_AUTHENTICATION_TAG_LENGTH, extractAesIvFromSharedSecret(sharedSecret));
    aesCipher.init(DECRYPT_MODE, secretKey, gcmParamSpec);

    return aesCipher.doFinal(encryptedData);
  }

  static X509Certificate generateSelfSignedX509Certificate(
      KeyPair keyPair,
      String dn,
      String signingAlgorithm,
      Date startDate,
      int certificateDurationDays
  ) {
    try {
      ContentSigner sigGen = new JcaContentSignerBuilder(signingAlgorithm)
          .setProvider(BOUNCY_CASTLE_PROVIDER)
          .build(keyPair.getPrivate());
      SubjectPublicKeyInfo subPubKeyInfo = SubjectPublicKeyInfo.getInstance(keyPair.getPublic().getEncoded());
      Date endDate = new Date(startDate.getTime() + TimeUnit.DAYS.toMillis(certificateDurationDays));

      X509v3CertificateBuilder certBuilder = new X509v3CertificateBuilder(
          new X500Name(dn),
          BigInteger.valueOf(1234),
          startDate, endDate,
          new X500Name(dn),
          subPubKeyInfo
      );

      JcaX509ExtensionUtils extensionUtils = new JcaX509ExtensionUtils();

      certBuilder.addExtension(
          Extension.keyUsage, true,
          new KeyUsage(KeyUsage.digitalSignature
              | KeyUsage.keyEncipherment
              | KeyUsage.dataEncipherment
              | KeyUsage.keyAgreement
              | KeyUsage.nonRepudiation
              | KeyUsage.cRLSign
              | KeyUsage.keyCertSign)
      );
      certBuilder.addExtension(Extension.basicConstraints, false, new BasicConstraints(true));
      certBuilder.addExtension(Extension.subjectKeyIdentifier, false, extensionUtils.createSubjectKeyIdentifier(keyPair.getPublic()));
      certBuilder.addExtension(Extension.authorityKeyIdentifier, false, extensionUtils.createAuthorityKeyIdentifier(keyPair.getPublic()));
      certBuilder.addExtension(Extension.extendedKeyUsage, false, new ExtendedKeyUsage(new KeyPurposeId[]{KeyPurposeId.id_kp_clientAuth, KeyPurposeId.id_kp_serverAuth}));
      X509CertificateHolder certificateHolder = certBuilder.build(sigGen);

      return new JcaX509CertificateConverter()
          .setProvider(BOUNCY_CASTLE_PROVIDER)
          .getCertificate(certificateHolder);
    } catch (CertificateException | CertIOException | NoSuchAlgorithmException | OperatorCreationException e) {
      throw new RuntimeException(e);
    }
  }

  static JWSHeader createJwsHeaderWith(JWSAlgorithm jwsAlgorithm, Base64URL sha256Thumbprint) {
    return new JWSHeader.Builder(jwsAlgorithm)
        .type(JWT)
        .x509CertSHA256Thumbprint(sha256Thumbprint)
        .build();
  }

  static JWTClaimsSet jwtClaimsSetFrom(Map<String, Object> claims) {
    JWTClaimsSet.Builder builder = new JWTClaimsSet.Builder();
    claims.forEach(builder::claim);

    return builder.build();
  }

  private static void validateRequestThumbprintAndSignature(SignedJWT requestJws, X509Certificate signingCertificate) {
    ZonedDateTime expiry = ZonedDateTime.ofInstant(signingCertificate.getNotAfter().toInstant(), systemDefault());
    if (!computeSHA256Thumbprint(signingCertificate).toString().equals(requestJws.getHeader().getX509CertSHA256Thumbprint().toString())
        || expiry.isBefore(ZonedDateTime.now())) {
      // NOTE: new certificates must be exchanged before the old ones expire. For a time two certificates will then be valid, the thumbprint determines which certificate to use for validation
      throw new RuntimeException("Thumbprint does not match or certificate is invalid.");
    }

    JWSVerifier jwsVerifier;
    if (signingCertificate.getPublicKey() instanceof ECPublicKey ecPublicKey) {
      try {
        jwsVerifier = new ECDSAVerifier(ecPublicKey);
      } catch (JOSEException e) {
        throw new RuntimeException("Failed to instantiate signature verifier.", e);
      }
    } else {
      jwsVerifier = new RSASSAVerifier((RSAPublicKey) signingCertificate.getPublicKey());
    }

    try {
      if (!requestJws.verify(jwsVerifier)) {
        throw new RuntimeException("The request signature is invalid.");
      }
    } catch (JOSEException e) {
      throw new RuntimeException("Failed to verify request signature.", e);
    }
  }

  private static KeyPair generateKeyPair(String algorithm, String parameterSpec, int keySize) {
    KeyPair keyPair;
    try {
      KeyPairGenerator keyPairGenerator = KeyPairGenerator.getInstance(algorithm, BOUNCY_CASTLE_PROVIDER);
      if (parameterSpec != null) {
        AlgorithmParameterSpec algorithmParameterSpec = getParameterSpec(parameterSpec);
        keyPairGenerator.initialize(algorithmParameterSpec);
      } else {
        keyPairGenerator.initialize(keySize);
      }

      keyPair = keyPairGenerator.generateKeyPair();
    } catch (InvalidAlgorithmParameterException | NoSuchAlgorithmException e) {
      throw new RuntimeException(e);
    }

    return keyPair;
  }

  private static byte[] extractAesKeyFromSharedSecret(byte[] sharedSecret) {
    return copyOfRange(sharedSecret, AES_KEY_SHARED_SECRET_OFFSET, AES_KEY_SHARED_SECRET_OFFSET + AES_KEY_BYTE_LENGTH);
  }

  private static byte[] extractAesIvFromSharedSecret(byte[] sharedSecret) {
    return copyOfRange(sharedSecret, AES_IV_SHARED_SECRET_OFFSET, AES_IV_SHARED_SECRET_OFFSET + AES_IV_BYTE_LENGTH);
  }

}
